# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import datetime
import hashlib
import os
import subprocess

import pytest

import lawwenda.datafiles
import lawwenda.lawwenda_cli
import lawwendatest


def test_list_shares():
    with lawwendatest.server() as svr:
        svr.cfg.add_share("/dev/null", name="foo", password=None)
        svr.cfg.add_share("/dev/null", name="bar", password=None)
        clishares = runcli(["--cfgpath", svr.cfg.path, "list_shares"])
        assert sorted([n for n in clishares.split("\n") if n]) == ["bar", "foo"]


def test_describe_share():
    with lawwendatest.server() as svr:
        svr.cfg.add_share("/dev/null", name="baz", password=None, include_by_tags=["foo"], title="bar")
        clidescshare = runcli(["--cfgpath", svr.cfg.path, "describe_share", "baz"])
        assert "foo" in clidescshare
        assert "bar" in clidescshare
        assert "baz" in clidescshare


def test_add_share():
    with lawwendatest.server() as svr:
        runcli(["--cfgpath", svr.cfg.path, "add_share", "sh1", "/dev/null"], inputdata="pwd\n")
        sh1 = svr.cfg.get_share_by_name("sh1")
        assert sh1.name == "sh1"
        assert sh1.path == "/dev/null"
        assert sh1.password_scrypt
        assert sh1.readonly
        assert not sh1.exclude_hidden
        daylater = datetime.datetime.now() + datetime.timedelta(days=1)
        runcli(["--cfgpath", svr.cfg.path, "add_share", "sh2", "/dev/null", "--title", "ttl",
                "--active-until", daylater.isoformat(), "--hide-by-pattern", "hp1", "--hide-by-pattern", "hp2",
                "--exclude-by-pattern", "ep", "--exclude-by-tag", "et", "--exclude-hidden", "--readwrite"],
               inputdata="\n")
        sh2 = svr.cfg.get_share_by_name("sh2")
        assert sh2.name == "sh2"
        assert sh2.path == "/dev/null"
        assert not sh2.password_scrypt
        assert not sh2.readonly
        assert sh2.exclude_hidden
        assert sh2.exclude_by_patterns == ["ep"]
        assert sh2.exclude_by_tags == ["et"]
        assert sorted(sh2.hide_by_patterns) == ["hp1", "hp2"]
        assert (sh2.active_until - daylater) < datetime.timedelta(minutes=1)


def test_remove_share():
    with lawwendatest.server() as svr:
        svr.cfg.add_share("/dev/null", name="baz", password=None, include_by_tags=["foo"], title="bar")
        assert svr.cfg.get_share_by_name("baz")
        runcli(["--cfgpath", svr.cfg.path, "remove_share", "baz"])
        assert svr.cfg.get_share_by_name("baz") is None


def test_generate_wsgi():
    with lawwendatest.server() as svr:
        assert runcli(["--cfgpath", svr.cfg.path, "generate_wsgi"]).strip() == svr.cfg.generate_wsgi().strip()


def test_run_dev_server():
    with lawwendatest.server() as svr:
        runcli(["--cfgpath", svr.cfg.path, "list_shares"])


def test_console_cfg():
    for addargs in [[], ["console"]]:
        with lawwendatest.server() as svr:
            clians = runcli(["--cfgpath", svr.cfg.path, *addargs],
                            inputdata="import hashlib\n"
                                      "print(hashlib.md5(cfg.path.encode()).hexdigest())\n")
            assert hashlib.md5(svr.cfg.path.encode()).hexdigest() in clians


def test_console_quickstart():
    separator = "#### foo bar baz ~~~~"
    with lawwendatest.server() as svr:
        clians = runcli(["--cfgpath", svr.cfg.path, "console"],
                        inputdata=f"print({separator!r})\n"
                                  f"quickstart()\n")
    ibegin = clians.find(separator)
    assert ibegin > -1
    assert "Quick start guide" in clians[ibegin:]


def test_console_readme():
    with lawwendatest.server() as svr:
        with pytest.raises(RuntimeError):
            runcli(["--cfgpath", svr.cfg.path, "console"], inputdata=f"import sys\nsys.exit(1)\n")
        runcli(["--cfgpath", svr.cfg.path, "console"], inputdata=f"import sys\ncallable(readme) or sys.exit(1)\n")


def runcli(args, *, inputdata: str = ""):  # TODO some stuff fails (timeout;password) when called from interactive shell
    #TODO again and again?!
    os.environ["PYTHONPATH"] = os.environ.get("PYTHONPATH", "") + ";" + lawwenda.datafiles.LAWWENDADIR
    p = subprocess.Popen([f"{lawwenda.datafiles.LAWWENDADIR}/lawwenda_cli.py", *args],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    ans = p.communicate(inputdata.encode(), timeout=60)[0].decode()
    if p.returncode != 0:
        raise RuntimeError("runcli failed.")
    return ans
