# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import base64
import contextlib
import io
import json
import urllib.error
import urllib.parse
import urllib.request
import zipfile

import pytest

import lawwenda.devserver
import lawwenda.fmapp
import lawwendatest


def test_root():
    with fmapp_server() as (share, app, svr, internals_url):
        with urllib.request.urlopen(svr.url) as f:
            assert b"</lawwenda-directoryview>" in f.read()


def test_endpoint_help():
    return  # TODO
    with fmapp_server() as (share, app, svr, internals_url):
        with urllib.request.urlopen(f"{internals_url}/help") as f:
            assert b"fritten" in f.read()


def test_endpoint_static():
    with fmapp_server() as (share, app, svr, internals_url):
        with urllib.request.urlopen(f"{internals_url}/static/backend/common.js") as f:
            assert b"export const backend" in f.read()


def test_endpoint_api_copy():
    with fmapp_server() as (share, app, svr, internals_url):
        app.filesystem.rootnode.child_by_name("d1").mkdir()
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        callapi(internals_url, "copy", srcpaths=["/f1"], destpath="/d1")
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        assert app.filesystem.rootnode.child_by_name("/d1/f1").is_file


def test_endpoint_api_delete():
    with fmapp_server() as (share, app, svr, internals_url):
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        callapi(internals_url, "delete", paths=["/f1"])
        assert not app.filesystem.rootnode.child_by_name("/f1").is_file


def test_endpoint_api_details():
    with fmapp_server() as (share, app, svr, internals_url):
        apianswer = callapi(internals_url, "details", path="/f1", method="GET")
        assert "mtime_ts" in json.loads(apianswer)


def test_endpoint_api_dir():
    with fmapp_server() as (share, app, svr, internals_url):
        apianswer = callapi(internals_url, "dir", path="/", config={}, method="GET")
        assert json.loads(apianswer)[0]["name"] == "f1"


def test_endpoint_api_known_tags():
    with fmapp_server() as (share, app, svr, internals_url):
        apianswer = callapi(internals_url, "known_tags", path="/f1", method="GET")
        assert isinstance(json.loads(apianswer), list)


def test_endpoint_api_mkdir():
    with fmapp_server() as (share, app, svr, internals_url):
        assert not app.filesystem.rootnode.child_by_name("d1").is_dir
        callapi(internals_url, "mkdir", path="/d1")
        assert app.filesystem.rootnode.child_by_name("d1").is_dir


def test_endpoint_api_move():
    with fmapp_server() as (share, app, svr, internals_url):
        app.filesystem.rootnode.child_by_name("d1").mkdir()
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        callapi(internals_url, "move", srcpaths=["/f1"], destpath="/d1")
        assert not app.filesystem.rootnode.child_by_name("/f1").is_file
        assert app.filesystem.rootnode.child_by_name("/d1/f1").is_file


def test_endpoint_api_rename():
    with fmapp_server() as (share, app, svr, internals_url):
        assert app.filesystem.rootnode.child_by_name("f1").is_file
        assert not app.filesystem.rootnode.child_by_name("f2").is_file
        callapi(internals_url, "rename", path="/f1", newname="f2")
        assert not app.filesystem.rootnode.child_by_name("f1").is_file
        assert app.filesystem.rootnode.child_by_name("f2").is_file


def test_endpoint_api_set_comment():
    return  # TODO
    with fmapp_server() as (share, app, svr, internals_url):
        assert not app.filesystem.rootnode.child_by_name("f1").comment
        callapi(internals_url, "set_comment", paths=["/f1"], comment="mycomment")
        assert app.filesystem.rootnode.child_by_name("f1").comment == "mycomment"


def test_endpoint_api_set_geo():
    pass  # TODO


def test_endpoint_api_set_rating():
    pass  # TODO


def test_endpoint_api_tag_entries():
    pass  # TODO


def test_endpoint_api_thumbnail():
    with fmapp_server() as (share, app, svr, internals_url):
        app.filesystem.rootnode.child_by_name("f.png").write_file(base64.b64decode(
            b"iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAIAAACQd1PeAAAAD0lEQVQIHQEEAPv/AP///wX+Av4DfRnGAAAAAElFTkSuQmCC"))
        apianswer = callapi(internals_url, "thumbnail", path="/f.png", method="GET")
        assert apianswer.startswith(b"\x89PNG\r\n\x1a\n")


def test_endpoint_api_untag_entries():
    pass  # TODO


def test_endpoint_api_upload():
    pass  # TODO


def test_endpoint_api_zip():
    with fmapp_server() as (share, app, svr, internals_url):
        apianswer = callapi(internals_url, "zip", paths=["/f1"])
        zipcont = callapi(svr.url + json.loads(apianswer)["url"], None, fullurl=True, method="GET")
        io.BytesIO(zipcont)
        with zipfile.ZipFile(io.BytesIO(zipcont)) as zf:
            assert zf.namelist() == ["f1"]


def test_auth():
    with fmapp_server(password="123") as (share, app, svr, internals_url):
        try:
            callapi(internals_url, "dir", path="/", config={}, method="GET")
            assert False
        except urllib.error.HTTPError as ex:
            assert ex.code == 401
        try:
            assert callapi(internals_url, "dir", path="/", config={}, method="GET", auth=("u", "321"))
            assert False
        except urllib.error.HTTPError as ex:
            assert ex.code == 403
        assert callapi(internals_url, "dir", path="/", config={}, method="GET", auth=("u", "123"))


def test_prevent_csrf():
    with fmapp_server() as (share, app, svr, internals_url):
        with pytest.raises(urllib.error.HTTPError):
            callapi(internals_url, "delete", paths=["/f1"], no_csrf=True)
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        with pytest.raises(urllib.error.HTTPError):
            callapi(internals_url, "delete", paths=["/f1"], wrong_csrf=True)
        assert app.filesystem.rootnode.child_by_name("/f1").is_file
        callapi(internals_url, "delete", paths=["/f1"])
        assert not app.filesystem.rootnode.child_by_name("/f1").is_file


@contextlib.contextmanager
def fmapp_server(*, password=None):
    with lawwendatest.tempdir() as cfgdir:  # TODO dedup
        cfg = lawwenda.Configuration(cfgdir)
        with lawwendatest.tempdir() as sharedir:
            share = cfg.add_share(sharedir, name="foo", password=password, readonly=False)
            fmapp = lawwenda.fmapp.FmApp(share)
            fmapp.filesystem.rootnode.child_by_name("f1").write_file(b"f1")
            svr = lawwenda.devserver._run_dev_server_for_app(fmapp)
            yield share, fmapp, svr, f"{svr.url}{lawwenda.fmapp.FmApp.URL_INTERNALS_NAME}"  # TODO noh named tuple ?!
            svr.shutdown()  # TODO dedup


def callapi(internals_url, methodname, *, method="POST", fullurl=False, no_csrf=False, wrong_csrf=False,
            auth=None, **kwargs):
    qstr = ""
    if method == "GET":
        qstr = "?" + urllib.parse.urlencode(kwargs)
        kwargs = None
    if no_csrf:
        headers = {}
    elif wrong_csrf:
        headers = {"Cookie": "csrftoken=x", "X-CSRFToken": "y"}
    else:
        headers = {"Cookie": "csrftoken=x", "X-CSRFToken": "x"}
    if auth:
        headers["Authorization"] = "Basic " + base64.b64encode(":".join(auth).encode()).decode()
    req = urllib.request.Request(internals_url if fullurl else f"{internals_url}/api/{methodname}/{qstr}",
                                 data=json.dumps(kwargs).encode() if kwargs else None,
                                 headers={"Content-Type": "application/json", **headers}, method=method)
    with urllib.request.urlopen(req) as f:
        return f.read()
