# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

import lawwenda.fs.factory
import lawwenda.fs.local
import lawwendatest

# "a unit test should test one thing"... well, yes... good that 'one thing' is rather vague... :)


def test_nodes():
    for fs in myfilesystems():
        d1 = fs.rootnode.child_by_name("d1")
        d2 = fs.rootnode.child_by_name("d2")
        for d in [d1, d2]:
            d.mkdir()
        f11 = d1.child_by_name("f11")
        f22 = d2.child_by_name("f22")
        f23 = d2.child_by_name("f23")
        f11.write_file(b"f11")
        f22.write_file(b"f22")
        f23.write_file(b"f23")
        assert fs.rootnode.path == "/"
        assert fs.rootnode.name == ""
        assert fs.rootnode.parent_node is None
        assert len(fs.rootnode.child_nodes) == 2
        assert len([n for n in fs.rootnode.child_nodes if n.name == "d1"]) == 1
        assert len([n for n in fs.rootnode.child_nodes if n.name == "d2"]) == 1
        assert len(d1.child_nodes) == 1
        assert len(d2.child_nodes) == 2
        assert d1.name == "d1"
        for d in [d1, d2]:
            assert d.exists
            assert d.is_dir
            assert not d.is_file
            assert not d.is_link
        for f in [f11, f22, f23]:
            assert f.exists
            assert not f.is_dir
            assert f.is_file
            assert not f.is_link
        assert f11.path == "/d1/f11"
        assert f11.name == "f11"
        assert f11.dirpath == "/d1"
        assert f11.parent_node.path == "/d1"
        assert f11.parent_node.name == "d1"
        assert f22.parent_node.path == "/d2"
        assert f22.parent_node.name == "d2"
        with f11.read_file() as f:
            assert f.read() == b"f11"
        with f22.read_file() as f:
            assert f.read() == b"f22"
        with f23.read_file() as f:
            assert f.read() == b"f23"
        d2.delete()
        for f in [d2, f22, f23]:
            assert not f.exists
            assert not f.is_dir
            assert not f.is_file
            assert not f.is_link
        assert d1.exists
        assert d1.is_dir
        assert not d2.is_file
        assert not d2.is_link
        assert f11.exists
        assert not f11.is_dir
        assert f11.is_file
        assert not f11.is_link


def test_nodes_metadata():
    fs = lawwendatest.MockFilesystem()
    f1 = fs.rootnode.child_by_name("f1")
    f1.write_file(b"")
    assert not f1.comment
    assert f1.rating == 0
    assert not f1.geo
    assert len(f1.tags) == 0
    f1.set_comment("foo")
    f1.set_rating(8)
    f1.set_geo("lon=1&lat=2")
    f1.add_tag("tag1")
    f1.add_tag("tag2")
    f1.add_tag("tag3")
    f1.remove_tag("tag2")
    assert f1.comment == "foo"
    assert f1.rating == 8
    assert ("lon=1" in f1.geo) and ("lat=2" in f1.geo)
    assert f1.geo_obj.lon() == 1
    assert f1.geo_obj.lat() == 2
    assert f1.tags == ["tag1", "tag3"]


def test_readonly():
    origfs = lawwendatest.MockFilesystem()
    rwfs1 = lawwenda.fs.factory.decorate_filesystem(origfs, readonly=True)
    rwfs2 = lawwenda.fs.factory.decorate_filesystem(rwfs1, readonly=False, exclude_by_tags=["dummy"])
    rwfs3 = lawwenda.fs.factory.decorate_filesystem(rwfs2, readonly=False, hide_by_tags=["dummy"])
    origfs.rootnode.child_by_name("f1").write_file(b"x")
    for rwfs in [rwfs1, rwfs2, rwfs3]:
        with pytest.raises(PermissionError):
            rwfs.rootnode.child_by_name("f1").write_file(b"y")
    for fs in [origfs, rwfs1, rwfs2, rwfs3]:
        assert fs.rootnode._filesystem == fs
        assert len(fs.rootnode.child_nodes) > 0 and all([n._filesystem == fs for n in fs.rootnode.child_nodes])
        assert fs.rootnode.child_by_name("x")._filesystem == fs
        with fs.rootnode.child_by_name("f1").read_file() as f:
            assert f.read() == b"x"


def test_exclude_by():
    origfs = lawwendatest.MockFilesystem()
    exfs1 = lawwenda.fs.factory.decorate_filesystem(origfs, exclude_by_tags=["extag"])
    exfs2 = lawwenda.fs.factory.decorate_filesystem(exfs1, exclude_by_patterns=[".*/ex.*"])
    exfs3 = lawwenda.fs.factory.decorate_filesystem(exfs2, hide_by_tags=["dummy"])
    exfs4 = lawwenda.fs.factory.decorate_filesystem(exfs3, readonly=True)
    d1 = origfs.rootnode.child_by_name("d1")
    d1.mkdir()
    f1 = d1.child_by_name("f1")
    f2 = d1.child_by_name("f2")
    f3 = d1.child_by_name("ex-f3")
    for f in [f1, f2, f3]:
        f.write_file(b"x")
    f1.add_tag("fu")
    f2.add_tag("extag")
    for fs in [exfs2, exfs3, exfs4]:
        assert [c.name for c in fs.rootnode.child_by_name("d1").child_nodes] == ["f1"]
        assert fs.rootnode.child_by_name("d1").child_by_name("f1").exists
        with pytest.raises(PermissionError):
            _ = fs.rootnode.child_by_name("d1").child_by_name("f2").exists
        with pytest.raises(PermissionError):
            _ = fs.rootnode.child_by_name("d1").child_by_name("ex-f3").exists
    for fs in [origfs, exfs1, exfs2, exfs3, exfs4]:
        assert fs.rootnode._filesystem == fs
        assert len(fs.rootnode.child_nodes) > 0 and all([n._filesystem == fs for n in fs.rootnode.child_nodes])
        assert fs.rootnode.child_by_name("x")._filesystem == fs


def test_include_by_tags():
    origfs = lawwendatest.MockFilesystem()
    infs1 = lawwenda.fs.factory.decorate_filesystem(origfs, include_by_tags=["in"])
    infs2 = lawwenda.fs.factory.decorate_filesystem(infs1, readonly=True, exclude_by_tags=["dummy"])
    infs3 = lawwenda.fs.factory.decorate_filesystem(infs2, readonly=True, hide_by_tags=["dummy"])
    d1 = origfs.rootnode.child_by_name("d1")
    d1.mkdir()
    d2 = origfs.rootnode.child_by_name("d2")
    d2.mkdir()
    f1 = d1.child_by_name("f1")
    f1.write_file(b"x")
    f1.add_tag("in")
    f2 = d2.child_by_name("f2")
    f2.write_file(b"x")
    f2.add_tag("fuh")
    assert sorted([c.name for c in origfs.rootnode.child_nodes]) == ["d1", "d2"]
    for infs in [infs1, infs2, infs3]:
        assert [c.name for c in infs.rootnode.child_nodes] == ["d1"]
        assert [c.name for c in infs.rootnode.child_by_name("d1").child_nodes] == ["f1"]
        assert infs.rootnode.child_by_name("d1").child_by_name("f1").exists
        with pytest.raises(PermissionError):
            _ = infs.rootnode.child_by_name("d2").exists


def test_include_by_pattern():
    origfs = lawwendatest.MockFilesystem()
    infs1 = lawwenda.fs.factory.decorate_filesystem(origfs, include_by_patterns=[".*in"])
    infs2 = lawwenda.fs.factory.decorate_filesystem(infs1, readonly=True, exclude_by_tags=["dummy"])
    infs3 = lawwenda.fs.factory.decorate_filesystem(infs2, readonly=True, hide_by_tags=["dummy"])
    d1 = origfs.rootnode.child_by_name("d1")
    d1.mkdir()
    d2 = origfs.rootnode.child_by_name("d2")
    d2.mkdir()
    d1.child_by_name("f1in").write_file(b"x")
    d2.child_by_name("f2out").write_file(b"x")
    assert sorted([c.name for c in origfs.rootnode.child_nodes]) == ["d1", "d2"]
    for infs in [infs1, infs2, infs3]:
        assert [c.name for c in infs.rootnode.child_nodes] == ["d1"]
        assert [c.name for c in infs.rootnode.child_by_name("d1").child_nodes] == ["f1in"]
        assert infs.rootnode.child_by_name("d1").child_by_name("f1in").exists
        with pytest.raises(PermissionError):
            _ = infs.rootnode.child_by_name("d2").exists


def test_hide_by():
    origfs = lawwendatest.MockFilesystem()
    hidefs1 = lawwenda.fs.factory.decorate_filesystem(origfs, hide_by_tags=["hideme"])
    hidefs2 = lawwenda.fs.factory.decorate_filesystem(hidefs1, readonly=True, exclude_by_tags=["dummy"])
    hidefs3 = lawwenda.fs.factory.decorate_filesystem(hidefs2, readonly=True, hide_by_tags=["dummy"])
    f1 = origfs.rootnode.child_by_name("f1")
    f1.write_file(b"x")
    f1.add_tag("barr")
    f2 = origfs.rootnode.child_by_name("f2")
    f2.write_file(b"x")
    f2.add_tag("hideme")
    assert sorted([c.name for c in origfs.rootnode.child_nodes]) == ["f1", "f2"]
    for infs in [hidefs1, hidefs2, hidefs3]:
        assert not infs.rootnode.child_by_name("f1").is_hidden
        assert infs.rootnode.child_by_name("f2").is_hidden
    exclhiddenfs = lawwenda.fs.factory.decorate_filesystem(hidefs3, exclude_hidden=True)
    assert exclhiddenfs.rootnode.child_by_name("f1").exists
    with pytest.raises(PermissionError):
        _ = exclhiddenfs.rootnode.child_by_name("f2").exists


def myfilesystems():
    yield lawwendatest.MockFilesystem()
    with lawwendatest.tempdir() as tmpdir:
        yield lawwenda.fs.local.LocalFilesystem(tmpdir)

# TODO basic features, write attempts ask for writehandle, read attempts ask for readhandle, decorators(also: correct filesystem in nodes)
