# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import contextlib
import datetime
import os

import lawwenda.fs.local
import lawwendatest


def test_is_file_link_dir():
    with filesystem() as (fs, rootpath):
        d1 = fs.rootnode.child_by_name("d1")
        assert d1.exists
        assert d1.is_dir
        assert not d1.is_file
        assert not d1.is_link
        f1 = d1.child_by_name("f1")
        assert f1.exists
        assert not f1.is_dir
        assert f1.is_file
        assert not f1.is_link
        ld1 = fs.rootnode.child_by_name("ld1")
        os.symlink(d1.try_get_fullpath(writable=False), ld1.try_get_fullpath(writable=True))
        assert ld1.exists
        assert ld1.is_dir
        assert not ld1.is_file
        assert ld1.is_link
        lf1 = fs.rootnode.child_by_name("lf1")
        os.symlink(f1.try_get_fullpath(writable=False), lf1.try_get_fullpath(writable=True))
        assert lf1.exists
        assert not lf1.is_dir
        assert lf1.is_file
        assert lf1.is_link
        l2 = fs.rootnode.child_by_name("l2")
        os.symlink(d1.child_by_name("notthere").try_get_fullpath(writable=False), l2.try_get_fullpath(writable=True))
        assert l2.exists
        assert not l2.is_dir
        assert not l2.is_file
        assert l2.is_link
        fn = fs.rootnode.child_by_name("fn")
        assert not fn.exists
        assert not fn.is_dir
        assert not fn.is_file
        assert not fn.is_link


def test_size():
    with filesystem() as (fs, rootpath):
        assert fs.rootnode.child_by_name("d1").child_by_name("f1").size == 2


def test_mtime():
    with filesystem() as (fs, rootpath):
        assert abs((fs.rootnode.child_by_name("d1").mtime - datetime.datetime.now()).total_seconds()) < 60 * 60


def test_metadata():
    pass # TODO


def test_child_nodes():
    with filesystem() as (fs, rootpath):
        assert [n.name for n in fs.rootnode.child_nodes] == ["d1"]
        assert [n.name for n in fs.rootnode.child_by_name("d1").child_nodes] == ["f1"]


def test_read_file():
    with filesystem() as (fs, rootpath):
        with fs.rootnode.child_by_name("d1").child_by_name("f1").read_file() as f:
            assert f.read() == b"f1"


def test_delete():
    with filesystem() as (fs, rootpath):
        assert os.path.exists(f"{rootpath}/d1")
        fs.rootnode.child_by_name("d1").delete()
        assert not os.path.exists(f"{rootpath}/d1")


def test_mkdir():
    with filesystem() as (fs, rootpath):
        assert not os.path.isdir(f"{rootpath}/d2")
        fs.rootnode.child_by_name("d2").mkdir()
        assert os.path.isdir(f"{rootpath}/d2")


def test_copy_to():
    with filesystem() as (fs, rootpath):
        fs.rootnode.child_by_name("d1").copy_to("D1")
        with open(f"{rootpath}/D1/f1", "rb") as f:
            assert f.read() == b"f1"
        with open(f"{rootpath}/d1/f1", "rb") as f:
            assert f.read() == b"f1"


def test_move_to():
    with filesystem() as (fs, rootpath):
        fs.rootnode.child_by_name("d1").move_to("D1")
        with open(f"{rootpath}/D1/f1", "rb") as f:
            assert f.read() == b"f1"
        assert not os.path.exists(f"{rootpath}/d1")


def test_write_file():
    with filesystem() as (fs, rootpath):
        assert not os.path.exists(f"{rootpath}/f2")
        fs.rootnode.child_by_name("f2").write_file(b"f2")
        with open(f"{rootpath}/f2", "rb") as f:
            assert f.read() == b"f2"


@contextlib.contextmanager
def filesystem():
    with lawwendatest.tempdir() as fsdir:
        os.mkdir(f"{fsdir}/d1")
        with open(f"{fsdir}/d1/f1", "wb") as f:
            f.write(b"f1")
        yield lawwenda.fs.local.LocalFilesystem(fsdir), fsdir


# TODO graceful degradation when some files are inaccessible to the process, no access outside of share path, no break out with symlinks,
