# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import contextlib
import subprocess
import urllib.error
import urllib.parse
import urllib.request

import lawwenda.sharehttpapp
import lawwendatest


def test_method_get():
    with sharehttpapp_server() as (sharehttpapp, svr):
        assert calldav(svr.url, "d1/f1", "GET") == b"f1"


def test_method_head():
    pass  # TODO


def test_method_propfind():
    with sharehttpapp_server() as (sharehttpapp, svr):
        assert b"/d1/f1" in calldav(svr.url, "", "PROPFIND")
        assert b"/d1/f1" not in calldav(svr.url, "", "PROPFIND", Depth="1")
        assert b"/d1/" in calldav(svr.url, "", "PROPFIND", Depth="1")
        assert b"/d1/" not in calldav(svr.url, "", "PROPFIND", Depth="0")
        assert calldav(svr.url, "", "PROPFIND") == calldav(svr.url, "", "PROPFIND", Depth="infinity")


def test_method_proppatch():
    pass  # TODO


def test_method_options():
    with sharehttpapp_server() as (sharehttpapp, svr):
        allows = [m.strip() for m in calldav(svr.url, "", "OPTIONS", return_headers=True)["Allows"].split(",")]
        assert "GET" in allows  # just some examples
        assert "PROPFIND" in allows
        assert "PUT" in allows


def test_method_put():
    pass  # TODO


def test_method_delete():
    with sharehttpapp_server() as (sharehttpapp, svr):
        assert sharehttpapp.filesystem.rootnode.child_by_name("d1").exists
        calldav(svr.url, "d1", "DELETE")
        assert not sharehttpapp.filesystem.rootnode.child_by_name("d1").exists


def test_method_mkcol():
    with sharehttpapp_server() as (sharehttpapp, svr):
        assert not sharehttpapp.filesystem.rootnode.child_by_name("d2").is_dir
        calldav(svr.url, "d2", "MKCOL")
        assert sharehttpapp.filesystem.rootnode.child_by_name("d2").is_dir


def test_method_copy():
    pass  # TODO


def test_method_move():
    pass  # TODO


def test_davclient_ls():
    with sharehttpapp_server() as (sharehttpapp, svr):
        answer = subprocess.Popen(["cadaver", svr.url], stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE).communicate(b"ls\n")
        assert " d1 " in answer[0].decode()


def test_davclient_put():
    with sharehttpapp_server() as (sharehttpapp, svr):
        with lawwendatest.tempdir() as tmpdir:
            with open(f"{tmpdir}/f2", "w") as f:
                f.write("f2")
            subprocess.Popen(["cadaver", svr.url], stdin=subprocess.PIPE, cwd=tmpdir).communicate(b"put f2\n")
            with sharehttpapp.filesystem.rootnode.child_by_name("f2").read_file() as f:
                assert f.read() == b"f2"


def test_davclient_get():
    with sharehttpapp_server() as (sharehttpapp, svr):
        with lawwendatest.tempdir() as tmpdir:
            subprocess.Popen(["cadaver", svr.url], stdin=subprocess.PIPE, cwd=tmpdir).communicate(b"get /d1/f1\n")
            with open(f"{tmpdir}/f1", "rb") as f:
                assert f.read() == b"f1"


@contextlib.contextmanager
def sharehttpapp_server():
    fs = lawwendatest.MockFilesystem()
    d1 = fs.rootnode.child_by_name("d1")
    d1.mkdir()
    f1 = d1.child_by_name("f1")
    f1.write_file(b"f1")
    sharehttpapp = lawwenda.sharehttpapp.ShareHttpApp(fs)
    svr = lawwenda.devserver._run_dev_server_for_app(sharehttpapp)
    yield sharehttpapp, svr
    svr.shutdown()  # TODO encapsulate nicer, TODO dedup


def calldav(davroot_url, path, method, *, requestxml=None, return_headers=False, **headers):  # TODO dedup
    req = urllib.request.Request(f"{davroot_url}{path}",
                                 data=requestxml,
                                 headers={"Content-Type": "text/xml", **headers}, method=method)
    with urllib.request.urlopen(req) as f:
        return f.headers if return_headers else f.read()  # TODO ugly


# TODO test csrf prevention
