# SPDX-FileCopyrightText: © 2021 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import contextlib
import dataclasses
import datetime
import io
import os
import shutil
import tempfile
import typing as t
import uuid

import lawwenda.devserver
import lawwenda.fs

if t.TYPE_CHECKING:
    import selenium.webdriver


class _TestServer:#TODO own config

    def __init__(self, url: str, cfg: lawwenda.Configuration):
        self.url = url
        self.cfg = cfg

    def add_share(self, path: str, *args, name: str = "shr1", password: str = "", **kwargs) -> lawwenda.Share:
        return self.cfg.add_share(path, *args, **kwargs, name=name, password=password)


class _TestTempServerShare:

    def __init__(self, svr: _TestServer, share: lawwenda.Share):
        self.server = svr
        self.__share = share

    @property
    def path(self) -> str:
        return self.__share.path

    @property
    def share_url(self) -> str:
        return f"{self.server.url}/{self.__share.name}/"


@contextlib.contextmanager
def tempdir() -> t.ContextManager[str]:
    ex_ = None
    for _ in range(100):
        try:
            ftmp = f"{tempfile.gettempdir()}/lawwenda-test-{uuid.uuid4()}"
            os.mkdir(ftmp)
            break
        except IOError as ex:
            ex_ = ex
    else:
        raise ex_
    try:
        yield ftmp
    finally:
        try:
            shutil.rmtree(ftmp)
        except IOError:
            pass


@contextlib.contextmanager
def server() -> t.ContextManager[_TestServer]:
    with tempdir() as tmp:
        cfg = lawwenda.Configuration(tmp)
        svr = lawwenda.devserver.run_dev_server(cfg)
        try:
            yield _TestServer(svr.url, cfg)
        finally:
            svr.shutdown()


@contextlib.contextmanager
def tempdir_server_share(**kwargs) -> t.ContextManager['_TestTempServerShare']:
    with tempdir() as tmp:
        with server() as svr:
            yield _TestTempServerShare(svr, svr.add_share(tmp, **kwargs))


@contextlib.contextmanager
def browser() -> t.ContextManager['selenium.webdriver.Firefox']:
    import selenium.webdriver
    driver = selenium.webdriver.Firefox()
    driver.implicitly_wait(30)
    driver.maximize_window()
    try:
        yield driver
    finally:
        driver.quit()


class MockFilesystem(lawwenda.fs.Filesystem):

    @dataclasses.dataclass
    class _Piece:
        name: str
        attrs: t.Dict[str, t.Any]
        mtime: datetime.datetime

    def __init__(self):
        super().__init__()
        self.__rootpiece = self._Piece("", {"dir": {}}, datetime.datetime.now())

    @staticmethod
    def __piece_by_parentpiece_and_name(parentpiece, name):
        for cpiece in (parentpiece.attrs.get("dir", {})).values():
            if cpiece.name == name:
                return cpiece
        return None

    def __piece_by_path(self, path):
        r = self.__rootpiece
        for pathsegm in [p for p in path.split("/") if p]:
            r = self.__piece_by_parentpiece_and_name(r, pathsegm)
            if not r:
                break
        return r

    def child_nodes(self, handle):
        result = []
        for cname in (self.__piece_by_path(handle.readable_node.path).attrs.get("dir", {}).keys()):
            result.append(handle.readable_node.child_by_name(cname))
        return result

    def is_hidden(self, handle):
        p = self.__piece_by_path(handle.readable_node.path)
        return p and ("hidden" in p.attrs)

    def is_dir(self, handle):
        p = self.__piece_by_path(handle.readable_node.path)
        return p and ("dir" in p.attrs)

    def is_file(self, handle):
        p = self.__piece_by_path(handle.readable_node.path)
        return p and ("file" in p.attrs)

    def is_link(self, handle):
        p = self.__piece_by_path(handle.readable_node.path)
        return p and ("link" in p.attrs)

    def exists(self, handle):
        return bool(self.__piece_by_path(handle.readable_node.path))

    def size(self, handle):
        return len(self.__piece_by_path(handle.readable_node.path).attrs.get("file", ""))

    def mtime(self, handle):
        return self.__piece_by_path(handle.readable_node.path).mtime

    def comment(self, handle):
        return self.__piece_by_path(handle.readable_node.path).attrs.get("comment", "")

    def rating(self, handle):
        return self.__piece_by_path(handle.readable_node.path).attrs.get("rating", 0)

    def tags(self, handle):
        return self.__piece_by_path(handle.readable_node.path).attrs.get("tags", [])

    def geo(self, handle):
        return self.__piece_by_path(handle.readable_node.path).attrs.get("geo", "")

    def delete(self, handle):
        parentpiece = self.__piece_by_path(handle.writable_node.parent_node.path)
        dircontent = parentpiece.attrs.get("dir", {})
        if handle.writable_node.name in dircontent:
            dircontent.pop(handle.writable_node.name)

    def mkdir(self, handle):
        parentpiece = self.__piece_by_path(handle.writable_node.parent_node.path)
        if not parentpiece:
            handle.writable_node.parent_node.mkdir()
        piece = self.__piece_by_path(handle.writable_node.path)
        if piece:
            if "dir" not in piece.attrs:
                raise OSError(f"{handle.writable_node.path} already exists")
        else:
            parentpiece.attrs["dir"][handle.writable_node.name] = self._Piece(handle.writable_node.name,
                                                                              {"dir": {}}, datetime.datetime.now())

    def copy_to(self, srchandle, desthandle):
        pass  # TODO

    def move_to(self, srchandle, desthandle):
        pass  # TODO

    def set_comment(self, handle, comment):
        self.__piece_by_path(handle.writable_node.path).attrs["comment"] = comment

    def set_geo(self, handle, geo):
        self.__piece_by_path(handle.writable_node.path).attrs["geo"] = lawwenda.fs.data.GeoLocation.from_geostring(geo)

    def set_rating(self, handle, rating):
        self.__piece_by_path(handle.writable_node.path).attrs["rating"] = rating

    def add_tag(self, handle, tag):
        p = self.__piece_by_path(handle.writable_node.path)
        tags = p.attrs["tags"] = p.attrs.get("tags", [])
        if tag not in tags:
            tags.append(tag)

    def remove_tag(self, handle, tag):
        p = self.__piece_by_path(handle.writable_node.path)
        tags = p.attrs["tags"] = p.attrs.get("tags", [])
        if tag in tags:
            tags.remove(tag)

    def read_file(self, handle):
        piece = self.__piece_by_path(handle.readable_node.path)
        if not piece:
            raise IOError(f"{handle.readable_node.path} does not exist")
        return io.BytesIO(piece.attrs["file"])

    def write_file(self, handle, content):
        parentpath = handle.writable_node.parent_node.path
        parentpiece = self.__piece_by_path(parentpath)
        if not parentpiece:
            raise IOError(f"{parentpath} does not exist")
        filepiece = parentpiece.attrs["dir"].get(handle.writable_node.name)
        if not filepiece:
            filepiece = self._Piece(handle.writable_node.name, {"file": b""}, datetime.datetime.now())
            parentpiece.attrs["dir"][handle.writable_node.name] = filepiece
        if not isinstance(content, bytes):
            content = content.read()
        filepiece.attrs["file"] = content
        filepiece.mtime = datetime.datetime.now()



# TODO some pylint for tests or not?
